/* rtapd by Vasim V. */
/* $Id: tnpppage.c,v 1.6 1999/04/28 09:41:10 vasim Exp $ */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include "crc.h"

#define MAX_PACKET_SIZE 4160
#define DEFAULT_SOCK_NAME "/var/run/rtnppd.queue.out"

int my_tnpp_addr;

char page_func = '@';


int tnpp_out;

int send_tnpp_packet(int fd, struct sockaddr_un *sun, unsigned char *data,
		     int len, int tnpp_to, int tnpp_from);

/* Send TNPP packet via unix socket */
int send_tnpp_packet(int fd, struct sockaddr_un *sun, unsigned char *data,
		      int len, int tnpp_to, int tnpp_from) {
  unsigned char buf[MAX_PACKET_SIZE];
  int length, crc;

  if (len > (MAX_PACKET_SIZE - 14))
    return;
  snprintf(buf, sizeof(buf) - 1, "%c%04X%02X%04X%02X%c", '\01', tnpp_to, 20,
	   tnpp_from, 0, '\02');
  length = strlen(buf);
  memcpy(buf + strlen(buf), data, len);
  length = length + len;
  buf[length] = '\03';
  length++;
  crc = crc16(buf, length);
  buf[length] = (unsigned char) (crc & 0xff);
  buf[length + 1] = (unsigned char) ((crc >> 8) & 0xff);
  length = length + 2;

  /* Send packet now */
  return (sendto(fd, buf, length, 0, (struct sockaddr *) sun,
		 strlen(sun->sun_path) + 2));
} /* send_tnpp_packet */  

void usage() {
  printf("\ntnpppage v1.2b by Vasim V.\n");
  printf("\nUsage:\ttnpppage [-q] -i PAGER-ID [-h ID-HIGH] [-p PAUSE] -t TNPP-ADDR [-s SOCKET] [-c FUNC] [-m MESSAGE] [-f MSGFILE]\n");
  printf("  or:\ttnpppage [-q] -l IDs-LIST [-p PAUSE] -t TNPP-ADDR [-s SOCKET] [-c FUNC] [-m MESSAGE] [-f MSGFILE]\n");
  printf("\nOptions:\n");
  printf("  -q\t\tWork quiet (print only error messages)\n");
  printf("  -i PAGER-ID\tSend message to pager PAGER_ID\n");
  printf("  -h ID-HIGH\tSend message to all pagers with id in range PAGER-ID .. ID-HIGH\n");
  printf("  -l IDs-LIST\tSend messages to pagers from list in file IDs-LIST (one id in one line, you can use symbols '#' and ';' as comments)\n");
  printf("  -t TNPP-ADDR\tSend message to device with address TNPP-ADDR\n");
  printf("  -p PAUSE\tWait PAUSE seconds after sending each message\n");
  printf("  -s SOCKET\tUse unix socket SOCKET (default is '%s')\n",
	 DEFAULT_SOCK_NAME);
  printf("  -c FUNC\tSend message with function code FUNC (default is 0x40)\n");
  printf("  -m MESSAGE\tSend message MESSAGE (default is read from stdin)\n");
  printf("  -f MSGFILE\tRead message from file MSGFILE\n\n");
} /* usage */

/* main */
int main(int argc, char *argv[]) {
  extern char *optarg;
  extern int optind;
  char *message = NULL;
  FILE *infile;
  char id[11];
  char *u_sock_name_out = DEFAULT_SOCK_NAME;
  char *msgfile;
  int dest = 0;
  int ch;
  unsigned char pbuf[MAX_PACKET_SIZE];
  unsigned char buf[4076];
  unsigned char *p;
  unsigned char *pc;
  char sbf[5];
  struct sockaddr_un u_sun;
  unsigned char c;
  int offset;
  int i;
  int res;
  int len;
  int silently;
  int id_from_list, id_range;
  char idhigh[11];
  char *filelist;
  char bufid[13];
  char bufline[128];
  int pause;

  id[0] = '\0';
  id_from_list = 0;
  id_range = 0;
  silently = 0;
  msgfile = NULL;
  pause = 0;
  /* Parse command-line options */
  while ((ch = getopt(argc, argv, "i:t:s:c:m:h:l:f:p:q")) != -1)
    switch (ch) {
    case 'i':
      strncpy(id, optarg, 11);
      id[10] = '\0';
      break;
    case 't':
      offset = 0;
      if ((*optarg == '0') && (strlen(optarg) > 1)
	&& (toupper(*(optarg + 1)) == 'X'))
	offset = 2;
      if (toupper(*optarg) == 'X')
	offset = 1;
      for (i = 0; (i < 4) && optarg[i + offset]; i++)
	sbf[i] = toupper(optarg[i + offset]);
      sbf[i] = '\0';
      dest = strtol(sbf, NULL, 16);
      break;
    case 's':
      u_sock_name_out = optarg;
      break;
    case 'c':
      offset = 0;
      if ((*optarg == '0') && (strlen(optarg) > 1)
	&& (toupper(*(optarg + 1)) == 'X'))
	offset = 2;
      if (toupper(*optarg) == 'X')
	offset = 1;
      for (i = 0; (i < 2) && optarg[i + offset]; i++)
	sbf[i] = toupper(optarg[i + offset]);
      sbf[i] = '\0';
      page_func = strtol(sbf, NULL, 16);
      break;
    case 'm':
      message = optarg;
      break;
    case 'p':
      pause = atoi(optarg);
      break;
    case 'f':
      msgfile = optarg;
      break;
    case 'h':
      strncpy(idhigh, optarg, 11);
      idhigh[10] = '\0';
      id_range = 1;
      break;
    case 'l':
      filelist = optarg;
      id_from_list = 1;
      break;
    case 'q':
      silently = 1;
      break;
    case '?':
      usage();
      exit(1);
    }

  if (!dest || (!id[0] && !id_from_list)) {
    fprintf(stderr, "\n\07Error! TNPP-address and pager id need!\n");
    usage();
    exit(1);
  }

  if (!message || msgfile) {
    if (msgfile) {
      infile = fopen(msgfile, "r");
      if (!infile) {
	fprintf(stderr, "Can't open message file\n");
	exit(2);
      }
    } else
      infile = stdin;
    res = fread(buf, 1, sizeof(buf), infile);
    if (res <= 0) {
      fprintf(stderr, "Can't read message\n");
      exit(2);
    }
    fclose(infile);
    p = buf;
  } else
    p = message;

  memset(&u_sun, 0, sizeof(u_sun));
  u_sun.sun_family = AF_UNIX;
  strncpy(u_sun.sun_path, u_sock_name_out, sizeof(u_sun.sun_path));
  tnpp_out = socket(AF_UNIX, SOCK_DGRAM, 0);
  if (tnpp_out < 0) {
    fprintf(stderr, "Error %s (%d) at creating unix socket\n",
	    strerror(errno), errno);
    exit(2);
  }

  pc = pbuf + 12;
  for (i = 0; (i < 4078) && (pc < (pbuf + sizeof(pbuf) - 2)); i++) {
    c = p[i];
    if (!c)
      break;
    if ((c != '\01') && (c != '\02') && (c != '\03') && (c != '\04')
	&& (c != '\05') && (c != '\06') && (c != '\025') && (c != '\027')
	&& (c != '\030') && (c != '\036') && (c != '\037') && (c != '\032')
	&& (c != '\015') && (c != '\012') && (c != '\033'))
      *pc++ = c;
    else {
      *pc++ = '\032';
      *pc++ = c + 0x40;
    }
  }

  if (id_from_list) {
    infile = fopen(filelist, "r");
    if (!infile) {
      fprintf(stderr, "Can't open file with list of IDs\n");
      exit(2);
    }
  }

  len = pc - pbuf - 12;
  while (1) {
    if (id_from_list) {
      do {
	p = fgets(bufline, sizeof(bufline) - 1, infile);
      } while (p && ((p[0] == '#') || (p[0] == ';')));
      if (!p)
	break;
      for (p = bufline; (*p && (*p != '\n') && (*p != ' ')
			 && (*p != ';') && (*p != '#')); p++);
      *p = '\0';
      if (!strlen(bufline))
	continue;
      strncpy(id, bufline, 11);
      id[10] = '\0';
    }

    sprintf(bufid, "B%c%- 10s", page_func, id);
    memcpy(pbuf, bufid, 12);
    if (!silently)
      printf("Sending message (%d chars) to pager %s (via TNPP-device %04X)...",
	     len, id, dest);

    /* Route to rtnppd */
    *pc = '\027';
    if (send_tnpp_packet(tnpp_out, &u_sun, pbuf, len + 12,
			 dest, 0) <= 0) {
      if (!silently)
	printf(" error!\n");
      fprintf(stderr, "Error '%s' (%d) at sending to unix socket message for '%s'\n",
	      strerror(errno), errno, id);
    } else
      if (!silently)
	printf(" ok!\n");
    if (id_range) {
      if (!strcmp(id, idhigh))
	break;
      for (p = id + strlen(id) - 1; p >= (unsigned char *) id; p--) {
	if ((*p >= '0') && (*p < '9')) {
	  *p = *p + 1;
	  break;
	} else
	  if (*p == '9')
	    *p = '0';
      }
    } else
      if (!id_from_list)
	break;
    if (pause) {
      if (!silently) {
	printf("Wait...\r");
	fflush(stdout);
      }
      sleep(pause);
    }
  }
  if (id_from_list)
    fclose(infile);
  close(tnpp_out);
  exit(0);
} /* main */
